#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Constants for the ZenML hub."""

ZENML_HUB_DEFAULT_URL = "https://hubapi.zenml.io/"
ZENML_HUB_ADMIN_USERNAME = "ZenML"
ZENML_HUB_CLIENT_VERIFY = False  # TODO: set to True once the hub has certs
ZENML_HUB_CLIENT_TIMEOUT = 10
ZENML_HUB_INTERNAL_TAG_PREFIX = "zenml-"
ZENML_HUB_VERIFIED_TAG = "zenml-badge-verified"
